
#include "VFAnalysis.h"

#include "Skeleton.h"

#include <math.h>


extern Polyhedron *poly;

unsigned char Integrator_opt = 0;  // the user specified integrator type (0 - Euler, 1 - RK2, 2 - RK4 ...)


/* -------------------------------------------------------------------------------------------
   The following are the routines for fixed point detection
*/

void capture_singular_tris(void)
{
	// for each triangle, determine whether there exists a fixed point or not by computing the winding number around its boundary
}


void compute_fixedpt_info(void)
{
	// for each triangle that contains a fixed point based on the previous computation
	// we compute its local Jacobian matrix by using the routine get_loc_Jacobian(...)

	// based on the obtained Jacobian of the triangle, which is also the Jacobian of the fixed point,
	// determine its type by solving the eigen values of the Jacobian matrix

	// if the fixed point is a saddle, also compute its two eigen vectors

	// store the result to a global list of fixed points
}


/*
    This function computes the local Jacobian matrix within a triangle that contains a fixed point.
	NOTE that this Jacobian is also the Jacobian of the fixed point.
*/
void get_loc_Jacobian(int tri, double Jacobian[2][2], double &x_cp, double &y_cp)
{
	double x[3], y[3], vx[3], vy[3];
	int i, j;
	Triangle *face;

	///These information has been obtained before
	///Can we reuse them and enhance the performance 
    face = poly->tlist[tri];

	//For each triangle, calculate the vector for all vertices
	for (j=0; j<face->nverts; j++) {
		x[j] = face->verts[j]->x;
		y[j] = face->verts[j]->y;

		vx[j] = face->verts[j]->vec.entry[0];
		vy[j] = face->verts[j]->vec.entry[1];
	}

	/////Second way to calculate the a b c d e f values

	double coord[3][3],  *inver_coord ;  //inver_coord[3][3];

	for(i = 0; i < 3; i++)
	{
		coord[0][i] = x[i];
		coord[1][i] = y[i];
		coord[2][i] = 1.;
	}

	inver_coord = MatrixOpp((double*)coord, 3, 3);

	icMatrix3x3 result, rightM;
	result.set(vx[0],vx[1],vx[2],  vy[0],vy[1],vy[2],  1,1,1);
	rightM.set(inver_coord[0], inver_coord[1], inver_coord[2],
				inver_coord[3], inver_coord[4], inver_coord[5],
				inver_coord[6], inver_coord[7], inver_coord[8]);


	result.rightMultiply(rightM);

	double a, b, c, d, e, f;

	a = result.entry[0][0];
	b = result.entry[0][1];
	c = result.entry[0][2];
	d = result.entry[1][0];
	e = result.entry[1][1];
	f = result.entry[1][2];

	//need to store it as a part of the information of the elements or unknown singularities
	Jacobian[0][0] = a;
	Jacobian[0][1] = b;
	Jacobian[1][0] = d;
	Jacobian[1][1] = e;

	//use to calculate the coordinates of the singularity 3/25/06
	x_cp = (f*b - c*e)/(a*e - d*b);
	y_cp = (c*d - a*f)/(a*e - b*d);
}


/*
   The following routine determines the type of a fixed point given its Jacobian
*/
int get_fixedPt_type(double jacobian[2][2])
{
	double a,b,d,e, r1, r2, i1, i2;

	a = jacobian[0][0];
	b = jacobian[0][1];
	d = jacobian[1][0];
	e = jacobian[1][1];

	//need to store it as a part of the information of the elements or unknow singularities

	//Getting the eigenvalue here

	double A, B, C, delta;

	A = 1;
	B = -(a + e);
	C = (a * e - b * d);

	delta =  B*B - 4*A*C;

	if( delta >= 0)
	{
		i1 = i2 = 0;
		r1 = (-B + sqrt(delta))/2;
		r2 = (-B - sqrt(delta))/2;
	}
	else
	{
		r1 = r2 = -B/2.;
		i1 = sqrt(-delta)/2;
		i2 = -sqrt(-delta)/2;
	}

	////////////////////////////////////////////////////////////////
	/// Finish the classification of the fixed point here!
}

/* --------------------------------------------------------------------------------------------
   The following are the routines for streamline tracing
*/


/*
   The following function computes the intersection between two line segmenets
   PointA and PointB form the first line segement
   PointC and PointD form the second line segment
   t: the parameter value of the intersection on line segment PointA and PointB
*/
int Trajectory::get_intersection(double PointA[2], double PointB[2], 
								 double PointC[2], double PointD[2], double t[2])
{

    double delta;
    double t1,t2;
    double a,b,c,d;
    double xba,yba,xdc,ydc,xca,yca;

    xba=PointB[0]-PointA[0];    yba=PointB[1]-PointA[1];
    xdc=PointD[0]-PointC[0];    ydc=PointD[1]-PointC[1];
    xca=PointC[0]-PointA[0];    yca=PointC[1]-PointA[1];

    delta=xba*ydc-yba*xdc;
    t1=xca*ydc-yca*xdc;
    t2=xca*yba-yca*xba;

    if(delta!=0)
    {
        t[0]=t1/delta;   t[1]=t2/delta;
        /*two segments intersect (including intersect at end points)*/
         if ( t[0]<=1 && (t[0]>=0 || fabs(t[0]) < 1e-8)
			 && t[1]<=1 && (t[1]>=0 || fabs(t[1]) < 1e-8 )) 
			 return 1;
       else return 0; 
    }

    else
    {       
        /* AB & CD are parallel. */
        if ( (t1!=0) && (t2!=0) ) return 2;

        /* when AB & CD are collinear */

        /*if AB isn't a vertical line segment, project to x-axis */
        if(PointA[0]!=PointB[0])   
        {
            a=min(PointA[0],PointB[0]); b=max(PointA[0],PointB[0]);
            c=min(PointC[0],PointD[0]); d=max(PointC[0],PointD[0]);

            if ( (d<a) || (c>b) ) return  3;
            else if( (d==a) || (c==b) ) return 4;  
            else return 5;
        }

        else         /* if AB is a vertical line segment, project to y-axis */  
        {

            a=min(PointA[1],PointB[1]); b=max(PointA[1],PointB[1]);
            c=min(PointC[1],PointD[1]); d=max(PointC[1],PointD[1]); 

            if( (d<a) || (c>b) ) return  3;
            else if( (d==a) || (c==b) ) return 4;
            else return 5;
        }
    }
}

/*
   The following function stores a temporary point list to the line segment 
   data structure of the trajectory / streamline.
*/
void Trajectory::store_to_global_line_segs(CurvePoints *temp, int num)
{
	int i;
	icVector2 dis_vec;

	LineSeg oneseg;

	for (i=0; i<num; i++)
	{
		oneseg.start.entry[0] = temp[i].gpx;
		oneseg.start.entry[1] = temp[i].gpy;

		oneseg.end.entry[0] = temp[i+1].gpx;
		oneseg.end.entry[1] = temp[i+1].gpy;

		////compute the length of this line segment
		dis_vec.entry[0] = temp[i+1].gpx - temp[i].gpx;
		dis_vec.entry[1] = temp[i+1].gpy - temp[i].gpy;

		oneseg.length = length(dis_vec);

		oneseg.Triangle_ID = temp[i].triangleid;

		// save to the global list
		linesegs.push_back(oneseg);
	}
}


/*
   The following function calls one of the integrator to compute the next point in streamline tracing.
   first: current point
   second: next position (both first and second are local coordinates)
   face_id: the triangle that the streamline is lying in now
   alaph: the barycentric coordiante of the current point
   type: 0 - forward tracing; 1 - backward tracing
   opt: 0 - Euler; 1 - RK2; 2 - others
*/

bool 
Trajectory::get_next_pt(double first[2], 
						double second[2], 
						int &face_id, 
						double alpha[3], 
						int type,
						unsigned char opt)
{
	switch (opt)
	{
	case 0:
		// Your Euler integrator
	case 1:
		// Your RK2 integrator
	case 2:
		// Your other integrator

	default:
		return false;
	}
}


/*
    The following function traces out the portion of a streamline within a given triangle and the entrance of the streamline.
	face_id : the triangle that the streamline is lying in now
	globalp : the starting point within the triangle, typically it is at one of the boundary edges of the triangle
	type : 0 - forward tracing; 1 - backward tracing
	flag:  1 - hit a fixed point; 2 - other reason to terminate a streamline
*/
int Trajectory::trace_in_triangle(int &face_id, double globalp[2], int type, int &flag)
{
	int i;
	double alpha[3];
	double cur_point[2], pre_point[2];
	double vert0[3];
	icVector3 VP, globalv;

	Triangle *face = poly->tlist[face_id];
	Triangle *pre_f = face;
	
	////Initialized the temporary point list

	CurvePoints *temp_point_list = (CurvePoints*) malloc(sizeof(CurvePoints) * 100);
	int NumPoints = 0;

	////get the local coordinates of the starting point

	VP.entry[0] = globalp[0] - face->verts[0]->x;
	VP.entry[1] = globalp[1] - face->verts[0]->y;
	VP.entry[2] = globalp[2] - face->verts[0]->z;

	// compute the local coordinate of the first point under the coordinate system of the current triangle
	// your later computation is performed under this coordinate system
	pre_point[0] = cur_point[0] = dot(VP, face->LX);
	pre_point[1] = cur_point[1] = dot(VP, face->LY);


	////////////////////////////////////////////////////
    for(i = 0; i < 100; i++)  
	{
		////1. calculate the barycentric coordinates for current point

		////2. if current point is inside current triangle
		////     store the current point to the temporary data structure, 
		////     then call the get_next_pt(...) routine to move on to the next point
		////   you can use the following code to store it
		/*
			temp_point_list[NumPoints].gpx = globalp[0];
			temp_point_list[NumPoints].gpy = globalp[1];
			temp_point_list[NumPoints].gpz = globalp[2];
			temp_point_list[NumPoints].lpx = cur_point[0];
			temp_point_list[NumPoints].lpy = cur_point[1];
			temp_point_list[NumPoints].triangleid = face->index;  
			NumPoints++;
		*/

		////3. Otherwise, the current point is outside of current triangle
		////   call the get_next_triangle(...) function to determine the next triangle
		////   that the streamline enters
		////   then break;

	}

	if(NumPoints > 1)
		store_to_global_line_segs(temp_point_list, NumPoints);

	free(temp_point_list);
	return face_id;
}


/*
    The following function determines the next triangle that the streamline will enter.
   face_id:  the triangle that the streamline is lying in now
   pre: the local coordinates of the previous integration point
   cur: the local coordinates of the current integration point
   param_t: the parameter value of the intersection of the streamline with a triangle edge if the streamline crosses an edge
   type: 0 - forward tracing; 1 - backward tracing
   PassVertornot: 0 - not pass a vertex; 1 - passes vertex0; 2 - passes vertex1; 3 - passes vertex2 of triangle "face_id"
   alpha: the barycentric coordinates of the current point
*/
void Trajectory::get_next_triangle(int &face_id, double pre[2], double cur[2], double param_t[2], int type, 
					 int &PassVertornot, double alpha[3])
{
	int which_edge = -1;

	int prev_face_id = face_id;

	Triangle *prev_face = poly->tlist[face_id];

	Vertex *vert = NULL;

	PassVertornot = 0;
	
	// First, we should see whether the streamline passes through a vertex or not
	// if it passes through a vertex, PassVertornot>0 and (PassVertornot-1) record which 
	// vertex the streamline passes. 
	// Specifically, 0 means vertex0 of this triangle, 1 means vertex1 of the current triangle, and so on.
	cross_a_vertex(face_id, cur, pre, type, PassVertornot);
	if(PassVertornot > 0)
	{
		return ;
	}

	// if the streamline does not pass through a vertex, then determine which edge it crosses.
	face_id = prev_face_id; 
	cross_boundary(pre, cur, face_id, alpha, which_edge, param_t);
	
	if(param_t[0] == -1 && param_t[1] == -1)
	{
		face_id = prev_face_id;   ////something wrong here
		return;
	}

	// According to the edge it passes, determine the next triangle it will enter
	pass_edge(face_id, which_edge);
}



/********************************************************************
If we have already judge which edge the curve will cross
We can use the edge information to get next triangle
face_id : the triangle that the streamline is lying in now
which_edge: 
********************************************************************/
void Trajectory::pass_edge(int &face_id, int which_edge)
{
	////Using edge information to get next triangle

	Triangle *face = poly->tlist[face_id];

	Edge *theone ;
	Vertex *v;

	for(int i = 0; i < 3; i++)
	{
		v = face->verts[which_edge];

		theone = face->edges[i];

		if(theone->verts[0] != v && theone->verts[1] != v)
			break;
	}

	Triangle *next_tri;
	if(theone->tris[0] != face)
		next_tri = theone->tris[0];
	else
		next_tri = theone->tris[1];

	if (theone->ntris<2)
		face_id = -1;
	else
		face_id = next_tri->index;

}

/*
   The function determines which triangle that the streamline passing through a specific vertex will enter.
   vert_id: the vertex id that the streamline passes through
   theone: the index of the next triangle
   type: 0 - forward tracing, 1 - backward tracing (you will need both for separatrices computation)
*/
void 
Trajectory::pass_vertex(int vert_id, int &theone, int type)
{
	Vertex *vert = poly->vlist[vert_id];

	int i;

	double vang;
	Corner *c;
	int NewTID = -1;

    int orient;

	////for 3D surfaces, we should not use the following angle directly!
	vang = atan2(vert->vec.entry[1], vert->vec.entry[0]);

	if(type == 1)
		vang += PI;

	if(vang < 0)
		vang += 2*PI;
	
	else if(vang > 2*PI)
		vang -= 2*PI;

	if(vert->corners[0]->orient)
		orient=1;
	else
		orient=0;

	for( i = 0; i < vert->ncorners; i++)
	{

		c = vert->corners[i];

		////first, we check which angle area the vector on the vertex will fall in
		if(orient > 0)
		{
			if(c->BeginAng > c->EndAng)
			{
				if((vang >= c->BeginAng && vang < 2*PI)|| (vang <= c->EndAng && vang >= 0))
				{
					NewTID = i;
					break;
				}
			}
			else{
				if(vang >= c->BeginAng && vang <= c->EndAng)
				{
					NewTID = i;
					break;
				}
			}
		}
		else{
			if(c->BeginAng < c->EndAng)
			{
				if((vang <= c->BeginAng && vang >= 0)|| (vang >= c->EndAng && vang < 2*PI))
				{
					NewTID = i;
					break;
				}
			}
			else{
				if(vang <= c->BeginAng && vang >= c->EndAng)
				{
					NewTID = i;
					break;
				}
			}
		}
	}

	if(NewTID < 0) //avoid crash!
	{
		theone = -1;
		return;
	}
	
	theone = vert->corners[NewTID]->t;
}


/*
   The following function determines whether the streamline passes through a vertex or not
   face_id:  the triangle that the streamline is lying in now
   pre_p: the local coordinates of the previous integration point
   cur_p: the local coordinates of the current integration point
   type: 0 - forward tracing; 1 - backward tracing
   passornot: 0 - not pass; 1 - passes vertex0; 2 - passes vertex1; 3 - passes vertex2 of triangle "face_id"
*/
void Trajectory::cross_a_vertex(int &face_id, double cur_p[2], double pre_p[2],int type, int &passornot)
{

	// determine whether there is a vertex of the current triangle falls on the line segment (pre_p, cur_p)

	// you can first determine a straight line that passes through pre_p and cur_p, 
	// then you determine whether there is a vertex satisfying this equation.

	// if there is a vertex satisfying such a straight line equation, 
	// call pass_vertex() function to determine the triangle that the streamline will enter
	// this next triangle will be one of the triangle that adjacent to the vertex that the streamline passes through

	// if none of the vertex satisfies the straight line equation, return
	passornot = 0;
}

/*
   If the streamline does not cross a vertex, then we determine which one of the edges 
   that it crosses.
   pre: the local coordinates of the previous integration point
   cur: the local coordinates of the current integration point
   face_id:  the triangle that the streamline is lying in now
   alpah: the barycentric coordinates of the current point
   which_edge: 0 - edge0 of the current triangle; 1 - edge1 of the current triangle; 2 - edge2
   t: the parameter of the intersection on the triangle edge
*/
void  Trajectory::cross_boundary(double pre[2], double cur[2], 
								 int face_id, double alpha[3], 
								 int &which_edge, double t[2])
{
	// determine which edge the streamline crosses based on the configuration of the barycentric coordiantes

	// use the get_intersection(...) function to determine whether the line segment (pre, cur) has valid
	// intersection with a triangle edge.
	// NOTE that this intersection computation has to be performed under the local coordiante system of the current triangle "face_id"
}


/*
   The following function computes a streamline from a given seed (with the global coordinates and the triangle that contains it)
   face_id: the index of the triangle that contains the seed
   x, y: the global or world coordinates of the seed
   type:  0 - forward tracing; 1 - backward tracing
*/
void Trajectory::cal_one_traj(int face_id, double x, double y, int type)
{
	int i;
	int flag = 0;
	double globalp[2];
	int pre_face, cur_face;

	pre_face = cur_face = face_id;

	globalp[0] = x;   globalp[1] = y;    

	// determine a maximum number of triangles that the streamline is allowed to pass
	int NUMTRACETRIS = (int)sqrt((double)poly->ntris);

	for(i = 0; i < 10*NUMTRACETRIS; i++)
	{

		if(cur_face == -1)  // terminate the computation if the streamline reaches the boundary of the domaim
		{
			return;
		}

		pre_face = cur_face;
		cur_face = trace_in_triangle(cur_face, globalp, type, flag); ////0 means always forward tracing here

		if(flag > 0 || pre_face == cur_face ) // if the streamline hits a fixed point or cannot leave a triangle, terminate it
		{
			return;
		}
	}
}

////////////////////////////////////////////////////////////////////////
////  Calculate separatrices starting from and ending at saddles
///

void  cal_seps()
{
	int i;

	// compute separatrices for each saddle.
	// NOTE that there are exactly four separatrices for each saddle: two outgoing that reach sink and two incoming ones that start from sources

}



/*----------------------------------------------------------------------*/
////Routines for matrix operations

/****************************************************
Maybe better for 2X2 and 3X3 matrix inversion
****************************************************/

double *MatrixOpp(double A[],int m,int n) //inverse 
{ 
    int i,j,x,y,k; 
    double *SP=NULL,*AB=NULL,*B=NULL,XX,*C; 
    SP=(double *)malloc(m*n*sizeof(double)); 
    AB=(double *)malloc(m*n*sizeof(double)); 
    B=(double *)malloc(m*n*sizeof(double)); 
    
    XX=Surplus(A,m,n); 
    XX=1/XX; 
    
    for(i=0;i<m;i++) 
	for(j=0;j<n;j++) 
	{
		for(k=0;k<m*n;k++) 
			B[k]=A[k]; 
			{
				for(x=0;x<n;x++) 
				B[i*n+x]=0; 
				for(y=0;y<m;y++) 
				B[m*y+j]=0; 
				B[i*n+j]=1; 
				SP[i*n+j]=Surplus(B,m,n); 
				AB[i*n+j]=XX*SP[i*n+j]; 
			} 
	} 

    C=MatrixInver(AB,m,n); 

	free(SP);
	free(AB);
	free(B);
    
    return C; 
} 
    
double * MatrixInver(double A[],int m,int n) //zhuanzhi
{ 
    int i,j; 
    double *B=NULL; 
    B=(double *)malloc(m*n*sizeof(double)); 
    
    for(i=0;i<n;i++) 
	for(j=0;j<m;j++) 
		B[i*m+j]=A[j*n+i]; 
    
    return B; 
} 
    
double Surplus(double A[],int m,int n) //hanglieshi
{ 
    
    int i,j,k,p,r; 
    double XX,temp=1,temp1=1,s=0,s1=0; 
    
    if(n==2) 
    {for(i=0;i<m;i++) 
    for(j=0;j<n;j++) 
    if((i+j)%2) temp1*=A[i*n+j]; 
    else temp*=A[i*n+j]; 
    XX=temp-temp1;} 
    else{ 
    for(k=0;k<n;k++) 
    {for(i=0,j=k;i<m,j<n;i++,j++) 
    temp*=A[i*n+j]; 
    if(m-i) 
    {for(p=m-i,r=m-1;p>0;p--,r--) 
    temp*=A[r*n+p-1];} 
    s+=temp; 
    temp=1; 
    } 
    
    for(k=n-1;k>=0;k--) 
    {for(i=0,j=k;i<m,j>=0;i++,j--) 
    temp1*=A[i*n+j]; 
    if(m-i) 
    {for(p=m-1,r=i;r<m;p--,r++) 
    temp1*=A[r*n+p];} 
    s1+=temp1; 
    temp1=1; 
    } 
    
    XX=s-s1;} 
    return XX; 
} 