
/*
VFAnalysis.h
This file provide routines fulfilling the analysis of 2D steady vector fields, including
fixed point detection, streamline computation, and topology computation.

Created by Guoning Chen 02/14/2013
*/

#pragma once

#ifndef __VFANALYSIS_H__
#define __VFANALYSIS_H__

#include "icVector.H"
#include "icMatrix.H"

#include <vector>
using namespace std;

/*
   The type of the fixed points
*/
enum SINGULARElEM{
	NOTHING,
	SOURCE,
	SINK,
	SADDLE,
	CWCENTER,
	CCWCENTER,
	REGULAR
};

class FixedPt{
public:
	bool connected;
	unsigned char type;                    //the type of the singularity
	int  index;                            //the index of the singularity in the list
	int  TriangleID;                       //the triangle containing the singularity
	int separatrices;                      //the index of the separatrix group belongs to the saddle 
	
	icVector2  gpos;                       //global coordinates of singularity
	icVector2  incoming, outgoing;         //the local eigen-vectors for saddle type fixed point
	icMatrix2x2 Jacobian;                  //the Jacobian associated with the fixed point

}; // end of FixedPt class


/* Line segment data structure to store the streamline */
typedef struct LineSeg{
	int Triangle_ID;            //which triangle this line segment locates in
	icVector2 start, end;    //local coordinates for start and end points
	double length;              //we may need to store the length of the line segment
} LineSeg;


/*For temporary trajetory*/ 
typedef struct CurvePoints{
	int triangleid;
	double lpx, lpy;
	double gpx, gpy, gpz;
	double length;
}CurvePoints;


class Trajectory{
public:
	int index;
	std::vector<LineSeg> linesegs;

	int saddleID;          /*which saddle this trajectory belongs to*/

	double eulerstep_scalar, RK4_stepsize;

	/*Construct the trajectory*/
	Trajectory(int index, int curMaxNum = 200);

	~Trajectory()
	{
		linesegs.clear();
	}

	void store_to_global_line_segs(CurvePoints *temp, int num);
	int get_intersection(double PointA[2], double PointB[2], double PointC[2], double PointD[2], double t[2]);
	int trace_in_triangle(int &face_id, double globalp[2], int type, int &flag);
	void get_next_triangle(int &face_id, double pre[2], double cur[2], double param_t[2], int type, 
					 int &PassVertornot, double alpha[3]);
	void cross_a_vertex(int &face_id, double cur_p[2], double pre_p[2],int type, int &passornot);
	void pass_edge(int &face_id, int which_edge);
	void pass_vertex(int vert_id, int &theone, int type);
    void cross_boundary(double pre[2], double cur[2], 
								 int face_id, double alpha[3], 
								 int &which_edge, double t[2]);

	void cal_one_traj(int face_id, double x, double y, int type);
	void cal_startpt_sep(int triangleID, icVector3 sep_vector, double saddle_ce[3], double newpos[3]);

	//// Here is the place to put your integrator

	bool get_next_pt(double first[2], double second[2], int &face_id, double alpha[3], int type, unsigned char opt);

	// Your Euler integrator

	// Your RK2 integrator

	// Your other integrator, say RK4, RK23, RK45


};  // end of Trajectory class


////singularities extraction
void capture_singular_tris(void);
void compute_fixedpt_info(void);
int get_fixedPt_type(double jacobian[2][2]);
void get_loc_Jacobian(int tri, double Jacobian[2][2], double &x_cp, double &y_cp);


/////functions for matrix inversion (you will need this to determine the type of a fixed point)
double *MatrixOpp(double A[],int m,int n); //inverse 
double *MatrixInver(double A[],int m,int n); //transpose
double Surplus(double A[],int m,int n); //


//// Separatrices computation
void  cal_seps();


#endif //__VFANALYSIS_H__